<?php

    /**
     * Generic base Vanilla layout class. It is meant to be extended by all
     * other layout classes. 
     */

    class NVM_Vanilla_Layout {

        /**
         * @var string The width of the layout. One of the following: narrow,
         *             normal, wide, full. 
         */

        public $width = 'normal';

        /**
         * @var bool Whether this layout is to be combined with the next one.
         */
        
        public $combine_with_next = false;

        /**
         * @var bool Whether this layout is compact, ie has no vertical margins
         *           and paddings.
         */
        
        public $compact = false;

        /**
         * @var array An array of classes to define the layout in HTML.
         */
        
        public $classes = array();

        /**
         * @var bool Whether the layout has a title or not.
         */
        
        public $has_title = false;

        /**
         * @var bool Whether the layout has a subtitle or not.
         */
        
        public $has_subtitle = false;

        /**
         * @var string The title of the layout.
         */
        
        public $title = null;

        /**
         * @var string The subtitle of the layout.
         */        
        
        public $subtitle = null;

        /**
         * @static
         * 
         * @var NVM_Vanilla_Layout Privately holds an internal instance of a 
         *                         generic layout to use in order to output a
         *                         layout's start and end HTML on demand.
         */
        
        private static $generic_layout = null;



        /**
         * Constructor of the layout.
         * 
         * @param array|null $layout_fields Array holding the initialisation
         *                                  fields of the layout. If left null
         *                                  then the fields will be attempted
         *                                  to be drawn from a surrounding ACF
         *                                  layout.
         */
        
        public function __construct( array $layout_fields=null ) {

            if ( $layout_fields == null ) {

                // Fields will be given by ACF layout row.

                $section_options         = get_sub_field( 'section_options' );
                $section_title           = get_sub_field( 'section_title' );

                $this->width             = $section_options['width'];
                $this->combine_with_next = $section_options['combine_with_next'];
                $this->compact           = $section_options['compact'];
                $this->classes           = array_merge( $this->classes, ! empty( $section_options['classes'] ) ? $section_options['classes'] : array() );

                $this->has_title         = $section_title['has_title'];
                $this->has_subtitle      = $section_title['has_subtitle'];
                $this->title             = $section_title['title'];
                $this->subtitle          = $section_title['subtitle'];

            } else {

                // Fields are given as function arguments. 

                if ( isset( $layout_fields['width'] ) ) {
                    $this->width = $layout_fields['width'];
                }

                if ( isset( $layout_fields['combine_with_next'] ) ) {
                    $this->combine_with_next = $layout_fields['combine_with_next'];
                }
                
                if ( isset( $layout_fields['compact'] ) ) {
                    $this->compact = $layout_fields['compact'];
                }

                $this->classes = array_merge( $this->classes, ! empty( $layout_fields['classes'] ) ? $layout_fields['classes'] : array() );
                
                if ( isset( $layout_fields['has_title'] ) ) {
                    $this->has_title = $layout_fields['has_title'];
                }
                
                if ( isset( $layout_fields['has_subtitle'] ) ) {
                    $this->has_subtitle = $layout_fields['has_subtitle'];
                }
                
                if ( isset( $layout_fields['title'] ) ) {
                    $this->title = $layout_fields['title'];
                }

                if ( isset( $layout_fields['subtitle'] ) ) {
                    $this->subtitle = $layout_fields['subtitle'];
                }

            }

        }



        /**
         * The function that outputs the HTML of the layout. Outputs nothing by
         * default and should be overriden in its subclasses.
         * 
         * @return void
         */

        public function layout() {}



        /**
         * Return the output of the layout's HTML as a string.
         * 
         * @return void
         */

        public function __toString () {

            ob_start();
            $this->layout();
            $output = ob_get_clean();

            return $output;
        }



        /**
         * @static
         * 
         * Privately initialises the static resources of the class.
         * 
         * @return void
         */

        private static function init () {

            if ( self::$generic_layout == null ) {
                self::$generic_layout = new NVM_Vanilla_Layout();
            }

        }



        /**
         * Outputs the opening HTML section of a layout.
         * 
         * @param array $extra_classes An array of extra classes to define the
         *                             layout in HTML.
         * 
         * @return void
         */

        public function section_start ( $extra_classes=array() ) { 

            $classes = 'vanilla-layout ' . 
                       'vanilla-layout-' . $this->width . ' ' . 
                       ( $this->combine_with_next ? ' vanilla-layout-combine-with-next' : '' ) .
                       ( $this->compact ? ' vanilla-layout-compact' : '' ) .
                       ( ! empty( $this->classes ) ? ' ' . implode( ' ', $this->classes ) : '' ) .
                       ( ! empty( $extra_classes ) ? ' ' . implode( ' ', $extra_classes ) : '' ); ?>

            <div class = "<?php echo $classes; ?>"> <?php

        }



        /**
         * @static
         * 
         * Is statically called in order to output the opening HTML section of a
         * generic layout.
         * 
         * @param array $extra_classes An array of extra classes to define the
         *                             layout in HTML.
         * 
         * @return void
         */

        public static function generic_section_start ( $extra_classes=array() ) { 

            self::init();
            self::$generic_layout->section_start( $extra_classes );

        }



        /**
         * Outputs the closing HTML section of a layout.
         * 
         * @return string
         */
        public function section_end () { ?>

            </div> <!-- .vanilla-layout --> <?php

        }



        /**
         * @static
         * 
         * Is statically called in order to output the closing HTML section of a
         * generic layout.
         * 
         * @param array $extra_classes An array of extra classes to define the
         *                             layout in HTML.
         * 
         * @return void
         */
        public static function generic_section_end () { 

            self::init();
            self::$generic_layout->section_end();

        }



        /**
         * Outputs the title part of the section of the layout.
         * 
         * @return void
         */
        public function section_title () { ?>

            <?php if ( $this->has_title || $this->has_subtitle ) : ?>
                <div class = "vanilla-layout-title clear">
                    <div class = "column text">
                        <?php if ( $this->has_title ) : ?>
                            <div class = "vanilla-layout-title-title">
                                <h2><?php echo $this->title; ?></h2>
                            </div>
                        <?php endif; ?> 
                        <?php if ( $this->has_subtitle ) : ?>
                            <div class = "vanilla-layout-title-subtitle">
                                <p><?php echo $this->subtitle; ?></p>
                            </div>
                        <?php endif; ?> 
                    </div>
                </div>
            <?php endif; ?> <?php

        }

    }

?>